from typing import Callable

from langchain_core.tools import StructuredTool, BaseTool


def gen_tool(name: str, desc: str, func: Callable[[any], any]) -> BaseTool:
    ans = StructuredTool.from_function(
        func=func,
        name=name,
        description=desc,
        # coroutine= ... <- you can specify an async method if desired as well
    )
    return ans



from pathlib import Path
from typing import List

from langchain_core.documents import Document

from langchain_cfg_build import langchain_cfg_build_app
from langchain_cfg_build.agent.agent_builder import AgentBuilder
from langchain_cfg_build.embed_db import embed_db_query_service
from langchain_cfg_build.embed_db.embed_db_loader import EmbedDbLoader
from langchain_cfg_build.llm.enum_llm import EnumLLM
from langchain_cfg_build.rag import rag_service
from langchain_cfg_build.tool import tool_utils


def generate_document_list() -> List[Document]:
    documents = [
        Document(page_content="Tom is a sunflower", metadata=rag_service.trim_metadata_dict(
            {"booktitle": "Magical summer rainy day", "location": {"x": 10, "y": 9}})),
        Document(page_content="BigCoCo is a water monster", metadata={"booktitle": "Magical summer rainy day"})
    ]
    # Add your document generation logic here
    return documents


def invoke_query(query: str):
    d_list = generate_document_list()
    db_path = '/tmp/dsa/embed_db/test'
    rag_service.save_embed_db(db_path, iter([d_list]))
    db_loader = EmbedDbLoader(local_path=db_path)
    resp = embed_db_query_service.retrieve_qa(db_loader=db_loader, llm=EnumLLM.gpt_4o_mini,
                                              query=query)
    return resp


if __name__ == '__main__':
    env_path = Path(__file__).parent.parent.parent
    app.initialize(str(env_path))
    # resp = invoke_query(query='BigCoCo 是什麼?')
    # print(resp)

    tool= tool_utils.gen_tool(name="Magical summer rainy day 小說資料庫",
                        desc="可以給定query,查詢Magical summer rainy day 小說的詳細資訊",
                        func=lambda x: invoke_query(query=x ))

    agent_builder = AgentBuilder(
        EnumLLM.gpt_4o,
        [],
        "hwchase17/react"
    )
    agent_builder.add_tool(tool)
    agent_executor = agent_builder.build_executor()
    query = '請用*Magical summer rainy day"* 書中的 *BigCoCo* 腳色,寫一個牠的故事'
    # query = 'What is the main focus of this project?'
    result = agent_executor.invoke({"input": query})
    print(result)



from dataclasses import dataclass

from langchain_chroma import Chroma
from langchain_openai import OpenAIEmbeddings


@dataclass
class EmbedDbLoader:
    local_path: str = None
    _db: Chroma = None

    def load(self) -> Chroma:
        if not self._db:
            self._db = self._load()
        return self._db

    def _load(self) -> Chroma:
        embeddings = OpenAIEmbeddings()
        if self.local_path:
            db = Chroma(persist_directory=self.local_path, embedding_function=embeddings)
            return db
        raise NotImplementedError(f"{self.__dict__} EmbedDbLoader load is not implemented!")



from langchain.chains.retrieval_qa.base import RetrievalQA

from langchain_cfg_build.embed_db.embed_db_loader import EmbedDbLoader
from langchain_cfg_build.llm.enum_llm import EnumLLM


def retrieve_qa(db_loader: EmbedDbLoader, llm: EnumLLM, query: str):
    db = db_loader.load()
    retriever = db.as_retriever(search_type="similarity", search_kwargs={"k": 5})
    chain = RetrievalQA.from_chain_type(llm=llm.value.get_instance(),
                                        chain_type='stuff',
                                        retriever=retriever,
                                        return_source_documents=True,
                                        verbose=True,
                                        input_key="question",
                                        )
    ans = chain.invoke(query)
    return ans
